/*!
    \file FeatureExtractor.cpp

    $Author: Frank Moosmann $ ( <moosmann@mrt.uka.de>)
    $Date: 2005/11/04 $
*/
 

#include "FeatureExtractor.h"
#include "FMUtils.h"

#ifdef HAVE_SIFT
# include "PgmIo.h"
# include "Filter.h"
# include "InterestPoint.h"
# include "ParametersManager.h"
# include "Normalizer.h"
# include "Detector.h"
# include "ResizeImage.h"
# include "Sift.h"
# define GLOBAL extern
# include "Parameters.h"
# undef GLOBAL
#endif

#include <cmath>
#include <iostream>
using namespace std;

/***************************************************************************
                         static functions
 ***************************************************************************/
string FeatureExtractor::asText(ExtractionMethod method)
{
  switch (method) {
    case ColorHSL:    return "HSL";
    case ColorRGB:    return "RGB";
    case Greyscale:   return "GREY";
    case WaveletHSL:  return "WaveletHSL";
    case WaveletRGB:  return "WaveletRGB";
    case WaveletGrey: return "WaveletGREY";
    case Gradient:    return "Gradient";
    case SIFT:        return "SIFT";
    case normSIFT:    return "normSIFT";
    case SIFTHSL:     return "SIFTHSL";
    case HSLplusSIFT: return "SIFT+meanHSL";
    default:          return "";
  }
}
FeatureExtractor::ExtractionMethod FeatureExtractor::asExtractionMethod (string method)
{
  if (method == "HSL")          return ColorHSL;
  if (method == "RGB")          return ColorRGB;
  if (method == "GREY")         return Greyscale;
  if (method == "WaveletHSL")   return WaveletHSL;
  if (method == "WaveletRGB")   return WaveletRGB;
  if (method == "WaveletGREY")  return WaveletGrey;
  if (method == "WAVELETHSL")   return WaveletHSL;
  if (method == "WAVELETRGB")   return WaveletRGB;
  if (method == "WAVELETGREY")  return WaveletGrey;
  if (method == "Gradient")     return Gradient;
  if (method == "GRADIENT")     return Gradient;
  if (method == "SIFT")         return SIFT;
  if (method == "normSIFT")     return normSIFT;
  if (method == "SIFTHSL")      return SIFTHSL;
  if (method == "SIFT+meanHSL") return HSLplusSIFT;
  return ColorHSL;
}

//**************************************************************************
//                        Constructors / Destructors
//**************************************************************************

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  \brief Constructor for FeatureExtractor

  A FeatureExtractor-Object must be instantiated for each image seperately.
  \param img the source image, already opened with Magick::ImageMagick
  \param pdfunction pointer to the probability density function object,
         which will be used to draw random subwindows. The PDF is also responsible
         to keep the constraint of maximum/minimum size of the subwindows
  \param method desired extraction Method
  \param maskFile string to the image in which the position of the object is defined
  \param maskFile string to the file containing the SIFT-Configuration
*/
FeatureExtractor::FeatureExtractor(Magick::Image img, std::string imgName, PDF* pdfunction, ExtractionMethod method, string siftCfgFile, string maskFile) {
  masterImg = img;
  imageName = imgName;
  pdf = pdfunction;
  extractionMethod = method;
  im_width = masterImg.columns();
  im_height = masterImg.rows();
  //debug_out(2,"image-width: %d, image-height: %d", (int)im_width, (int)im_height);
  
  // ******** Do preprocessing once per image ********
  integImg = 0;
#ifdef HAVE_SIFT
  imagePyr = 0;
  imagePyr2 = 0;
  imagePyr3 = 0;
#endif
  switch (extractionMethod) {
    case WaveletHSL:
      integImg = new IntegralImage(masterImg, IntegralImage::ColorHSL);
      break;
    case WaveletRGB:
      integImg = new IntegralImage(masterImg, IntegralImage::ColorRGB);
      break;
    case WaveletGrey:
      integImg = new IntegralImage(masterImg, IntegralImage::Greyscale);
      break;
    case Gradient:
      //masterImg.enhance();
      //masterImg.unsharpmask(3,2,80,0);
      //masterImg.blur(5,2.5);
      masterImg.edge(0.);
      break;
#ifdef HAVE_SIFT
    case SIFT:
    case normSIFT:
    case HSLplusSIFT:
      InitParameters(siftCfgFile.c_str());
      SiftScaleMultiplier = ((float)WINDOW_SIZE) / SiftWindowSize;
      Image *cImage;
      CreateCImage(&cImage);
      InitPyramid(cImage, &imagePyr);
      InitResizer();
      BuildPyramid(cImage, imagePyr);
      DeallocResizer();
      DeallocImage(cImage);
      InitNormalizer();
      SiftInit();
      if (extractionMethod == HSLplusSIFT)
        integImg = new IntegralImage(masterImg, IntegralImage::ColorHSL);
      break;
    case SIFTHSL:
      InitParameters(siftCfgFile.c_str());
      SiftScaleMultiplier = ((float)WINDOW_SIZE) / SiftWindowSize;
      Image *cImageH;
      Image *cImageS;
      Image *cImageL;
      CreateCImage(&cImageH, &cImageS, &cImageL);
      InitPyramid(cImageH, &imagePyr);
      InitPyramid(cImageS, &imagePyr2);
      InitPyramid(cImageL, &imagePyr3);
      InitResizer();
      BuildPyramid(cImageH, imagePyr);
      BuildPyramid(cImageS, imagePyr2);
      BuildPyramid(cImageL, imagePyr3);
      DeallocResizer();
      DeallocImage(cImageH);
      DeallocImage(cImageS);
      DeallocImage(cImageL);
      InitNormalizer();
      SiftInit();
      break;
#endif
    default: ;
  }
  // ******** Read the mask from file ********
  if (maskFile.compare("")!=0) {
    maskImg = new MaskImage(maskFile, im_width, im_height);
    relObjSize = maskImg->getObjBgrRatio(im_width, im_height, 0, 0);
  } else {
    maskImg = 0;
    relObjSize = pdf->relObjSize;
  }
  //cout << relObjSize << flush;
}

FeatureExtractor::~FeatureExtractor() {
  if (maskImg != 0)
    delete maskImg;
  if (integImg != 0)
    delete integImg;
#ifdef HAVE_SIFT
  if (imagePyr != 0) {
    DeallocNormalizer();
    SiftDealloc();
    DeallocPyramid(imagePyr);
  }
  if (imagePyr2 != 0)
    DeallocPyramid(imagePyr2);
  if (imagePyr3 != 0)
    DeallocPyramid(imagePyr3);
#endif
}


//**************************************************************************
//                        public Methods
//**************************************************************************


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  \brief Method to extract features from an image

  The following method randomly takes subwindows, and returns its features.
  These are defined by the extraction method given in FeatureExtractor()
  \param featureNumber number specifying how many features(subwindows) to extract
  \param featureList pointer to the list where the FestureVectors will be stored
         <br>ATTENTION: this method allocates memory for the feature vectors, 
         hence feature vectors must be freed later using "delete";
  \param minObjBgrRatio if a maskfile exists, only those subwindows will be used, that have a ratio
         between Object and Background higher than this number (0.0 = all subwindows will be used)
*/
void FeatureExtractor::extractFeatures(const unsigned int featureNumber, std::list<FeatureVector*>* featureList, double minObjBgrRatio)
{
  //cout << "extracting "<< featureNumber <<" features..." << flush;
  //featureList->clear();
  unsigned int curr_window_Size;
  unsigned int curr_window_xOffset;
  unsigned int curr_window_yOffset;
  unsigned int nbAcceptedFeatures = 0;
  while (nbAcceptedFeatures < featureNumber)
  { 
    // ****************  generate random window
    //unsigned int maxSize = (unsigned int)(((double)min(im_height,im_width) * maxSubwinSize / 100) - 1);
    //curr_window_Size = randomInRange(WINDOW_SIZE,maxSize);
    //curr_window_xOffset = randomInRange (0, im_width - curr_window_Size);
    //curr_window_yOffset = randomInRange (0, im_height - curr_window_Size);
    pdf->randomSample(&curr_window_xOffset, &curr_window_yOffset, &curr_window_Size);
    //curr_window_Size = max(min(im_width, im_height)-50, (unsigned int)WINDOW_SIZE);
    //curr_window_xOffset = 15;
    //curr_window_yOffset = 25;
    // ****************  test weather to keep the window
    if ((maskImg != 0) && (minObjBgrRatio > 0.0)) {
      double ratio = maskImg->getObjBgrRatio(curr_window_Size, curr_window_Size, curr_window_xOffset, curr_window_yOffset);
      if (ratio < minObjBgrRatio) {
        continue;
      }
    }

    // ****************  set feature vector
    FeatureVector* featureVec = new FeatureVector();
    featureVec->imageName = imageName;
    featureVec->imageWidth = im_width; 
    featureVec->imageHeight = im_height; 
    featureVec->windowPosX = curr_window_xOffset; 
    featureVec->windowPosY = curr_window_yOffset;
    featureVec->windowSize = curr_window_Size;
    if (relObjSize > 0)
      featureVec->windowToObjectRatio = (curr_window_Size*curr_window_Size) / (relObjSize*im_width*im_height);
    else
      featureVec->windowToObjectRatio = 0.0;
    featureVec->classLabel = 0;
    
    switch (extractionMethod) {
      case ColorHSL:
      case ColorRGB:
      case Greyscale:
      case Gradient: {
          Magick::Image second = masterImg;
          second.crop(Magick::Geometry(curr_window_Size,curr_window_Size,curr_window_xOffset,curr_window_yOffset)); // (width, height, xOffset, yOffset)
          second.scale(Magick::Geometry(WINDOW_SIZE,WINDOW_SIZE));      // Resize image by using simple ratio algorithm
          //second.zoom(Geometry(WINDOW_SIZE,WINDOW_SIZE));     // Zoom image to specified size
          //second.sample(Geometry(WINDOW_SIZE,WINDOW_SIZE));   // Resize image by using pixel sampling algorithm
          switch (extractionMethod) {
            case ColorHSL: extractHSLFeatures(second, WINDOW_SIZE, featureVec); break;
            case ColorRGB: extractRGBFeatures(second, WINDOW_SIZE, featureVec); break;
            case Greyscale:extractGreyFeatures(second, WINDOW_SIZE, featureVec); break;
            case Gradient: extractGreyFeatures(second, WINDOW_SIZE, featureVec); break;
            default: ;
          }
        }
        break;
      case WaveletHSL:
      case WaveletRGB:
      case WaveletGrey: {
          extractWaveletFeatures(curr_window_Size, curr_window_xOffset, curr_window_yOffset, WINDOW_SIZE, featureVec);
        }
        break;
 #ifdef HAVE_SIFT
      case SIFT:
      case SIFTHSL:
      case normSIFT: {
          extractSIFTFeatures(curr_window_Size, curr_window_xOffset, curr_window_yOffset, featureVec, (extractionMethod==normSIFT));
          if (featureVec->getSize() == 0) {
            delete featureVec; //cout << "x" << flush;
            continue;
          }
        }
        break;
      case HSLplusSIFT: {
          extractSIFTFeatures(curr_window_Size, curr_window_xOffset, curr_window_yOffset, featureVec, (extractionMethod==normSIFT));
          if (featureVec->getSize() == 0) {
            delete featureVec; //cout << "x" << flush;
            continue;
          }
          extractHSLBins(curr_window_Size, curr_window_xOffset, curr_window_yOffset, 8, featureVec);
        }
        break;
#endif
      default:;
    }
    //debug_out(10,featureVec);
    featureList->push_back(featureVec);
    nbAcceptedFeatures++;
  } // end of for (int featureNo=1; featureNo<=numberOfFeatures; featureNo++)
    
}


//***************************************************************************
//                       private Methods
//***************************************************************************
void FeatureExtractor::extractHSLFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec) {
  for (unsigned int row=0; row<windowSize; row++)  {
    for (unsigned int col=0; col<windowSize; col++)  {
      Magick::Color pixCol = window.pixelColor( col, row );
      Magick::ColorHSL hslPixCol = Magick::ColorHSL(pixCol);  // white: 0.0h 0.0s 1.0l, black: 0.0h 0.0s 0.0h
      featureVec->addElement(hslPixCol.hue());
      featureVec->addElement(hslPixCol.saturation());
      featureVec->addElement(hslPixCol.luminosity());
    }
  }
}

void FeatureExtractor::extractRGBFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec) {
  for (unsigned int row=0; row<windowSize; row++)  {
    for (unsigned int col=0; col<windowSize; col++)  {
      Magick::Color pixCol = window.pixelColor( col, row );
      Magick::ColorRGB rgbPixCol = Magick::ColorRGB(pixCol);  // white: 1.0r 1.0g 1.0b, black: 0.0r 0.0g 0.0b
      featureVec->addElement(rgbPixCol.red());
      featureVec->addElement(rgbPixCol.green());
      featureVec->addElement(rgbPixCol.blue());
    }
  }
}

void FeatureExtractor::extractGreyFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec) {
  for (unsigned int row=0; row<windowSize; row++)  {
    for (unsigned int col=0; col<windowSize; col++)  {
      featureVec->addElement((Magick::ColorGray(window.pixelColor(col,row))).shade());
    }
  }
}

void FeatureExtractor::extractWaveletFeatures(unsigned int subwindowSize, unsigned int xOffset, unsigned int yOffset, unsigned int targetSize, FeatureVector* featureVec) {
  unsigned int nbChannels = integImg->getNoChannels();
  for (unsigned int ch=0; ch<nbChannels; ch++) {
    featureVec->addElement(integImg->getIntegralArea(xOffset,yOffset,subwindowSize,subwindowSize,ch)/(subwindowSize*subwindowSize)); // add average
  }
  int level = (int)log2(targetSize);
  extractWaveletCoefficients(subwindowSize, subwindowSize, xOffset, yOffset, level, featureVec);
}

void FeatureExtractor::extractHSLBins(unsigned int subwindowSize, unsigned int xOffset, unsigned int yOffset, unsigned int nbBins, FeatureVector* featureVec) {
  unsigned int nbChannels = integImg->getNoChannels();
  //nbChannels -= 1; // hack: only hue+saturation, not luminosity
  vector<unsigned int> xPos; xPos.push_back(xOffset);
  vector<unsigned int> yPos; yPos.push_back(yOffset);
  for (unsigned int i=0; i<nbBins; i++) {
    unsigned int width_i = subwindowSize / (nbBins-i);
    xPos.push_back(xPos[i]+width_i);
    yPos.push_back(yPos[i]+width_i);
    subwindowSize -= width_i;
  }
  for (unsigned int x=0; x<nbBins; x++) {
    for (unsigned int y=0; y<nbBins; y++) {
      for (unsigned int ch=0; ch<nbChannels; ch++) {
        unsigned int sizeX = xPos[x+1]-xPos[x];
        unsigned int sizeY = yPos[y+1]-yPos[y];
        featureVec->addElement(integImg->getIntegralArea(xPos[x],yPos[y],sizeX,sizeY,ch)/(sizeX*sizeY)); // add average
      }
    }
  }
}

void FeatureExtractor::extractWaveletCoefficients(unsigned int width, unsigned int height, unsigned int posX, unsigned int posY, int level, FeatureVector* featureVec) {
  //cout << "  getting coefficients at level "<<level<< " for windows " <<posX<<","<<posY<<"/"<<width<<","<<height<<endl;
  if (level < 1)
    return;
  // T=top B=bottom, L=left, R=right
  unsigned int widthL = (unsigned int)(width/2);
  unsigned int widthR = width-widthL;
  unsigned int heightT = (unsigned int)(height/2);
  unsigned int heightB = height-heightT;
  unsigned int splitX = posX + widthL;
  unsigned int splitY = posY + heightT;
  
  unsigned int nbChannels = integImg->getNoChannels();
  for (unsigned int ch=0; ch<nbChannels; ch++) {
    double meanTL = integImg->getIntegralArea(posX,posY,widthL,heightT,ch) / (widthL*heightT);
    double meanTR = integImg->getIntegralArea(splitX,posY,widthR,heightT,ch) / (widthR*heightT);
    double meanBL = integImg->getIntegralArea(posX,splitY,widthL,heightB,ch) / (widthL*heightB);
    double meanBR = integImg->getIntegralArea(splitX,splitY,widthR,heightB,ch) / (widthR*heightB);
    featureVec->addElement((-meanTL -meanTR +meanBL +meanBR));
    featureVec->addElement((+meanTL -meanTR +meanBL -meanBR));
    featureVec->addElement((-meanTL +meanTR +meanBL -meanBR));
  }
  level--;
  extractWaveletCoefficients(widthL, heightT, posX, posY, level, featureVec);
  extractWaveletCoefficients(widthR, heightT, splitX, posY, level, featureVec);
  extractWaveletCoefficients(widthL, heightB, posX, splitY, level, featureVec);
  extractWaveletCoefficients(widthR, heightB, splitX, splitY, level, featureVec);
}


#ifdef HAVE_SIFT
void FeatureExtractor::extractSIFTFeatures(unsigned int subwindowSize, unsigned int xOffset, unsigned int yOffset, FeatureVector* featureVec, bool normalize) {
  //cout << "E" << flush;
  InterestPoint *pt = new InterestPoint;
  pt->scale = (float)subwindowSize/(float)WINDOW_SIZE;
  pt->x = xOffset+(subwindowSize/2);
  pt->y = yOffset+(subwindowSize/2);
  pt->scaleIndex = 1; // not used any more
  pt->mi[0] = 1;
  pt->mi[1] = 0;
  pt->mi[2] = 0;
  pt->mi[3] = 1;
  pt->theta = 0;
  pt->descr = 0;
  pt->next = 0;
  if (normalize) {
    NormalizeInterestPoints(imagePyr, &pt);
    CleanPoints(&pt);
  }
  if (pt != 0) {
    SiftDescribe(imagePyr, pt);
    int fcount = SiftIndexSize*SiftIndexSize*SiftOriSize;
    for (int i=0; i<fcount; i++) {
      FeatureData val = pt->descr[i];
      if (isnan(val))
        cerr << "Juliens code is not nice with image " << imageName << " at x/y/s " << xOffset << "/" << yOffset << "/" << subwindowSize;
      else
        featureVec->addElement(val);
    }
    if (imagePyr2 != 0) {
      SiftDescribe(imagePyr2, pt);
      for (int i=0; i<fcount; i++) {
        featureVec->addElement(pt->descr[i]);
      }
    }
    if (imagePyr3 != 0) {
      SiftDescribe(imagePyr3, pt);
      for (int i=0; i<fcount; i++) {
        featureVec->addElement(pt->descr[i]);
      }
    }
    DeallocPoints(pt);
  }
  //cout << featureVec->getSize() << endl;
}

void FeatureExtractor::CreateCImage(Image **pcImage, Image **pcImage2, Image **pcImage3) {
  bool color = ((pcImage2 != 0) && (pcImage3 != 0));
  unsigned int imWidth = masterImg.columns();
  unsigned int imHeight = masterImg.rows();
//masterImg.display();
  Image *cImage;
  Image *cImage2 = 0;
  Image *cImage3 = 0;
  AllocImage(imWidth, imHeight, 255, pcImage); cImage = *pcImage;
  if (color) {
    AllocImage(imWidth, imHeight, 255, pcImage2); cImage2 = *pcImage2;
    AllocImage(imWidth, imHeight, 255, pcImage3); cImage3 = *pcImage3;
  }
  for (unsigned int row=0; row<imHeight; row++) {
    for (unsigned int col=0; col<imWidth; col++) {
      if (color) {
        Magick::ColorHSL hslPixCol = Magick::ColorHSL(masterImg.pixelColor(col-1,row-1));
        cImage->data[row][col] = (float)(hslPixCol.hue());
        cImage2->data[row][col] = (float)(hslPixCol.saturation());
        cImage3->data[row][col] = (float)(hslPixCol.luminosity());
      } else
        cImage->data[row][col] = (float)Magick::ColorGray(masterImg.pixelColor(col-1,row-1)).shade();
    }
  }
//SavePgm("test.pgm", cImage, 0);
//masterImg.display();
}
#endif /* HAVE_SIFT */

// end FeatureExtractor.cpp
