/***************************************************************************
                                  PDF.cpp
                             ------------------
    begin                : 07.12.2005
    copyright            : (C) 2005 by Frank Moosmann and INRIA
    email                : moosmann@mrt.uka.de
 ***************************************************************************/
//#define TESTHACK 32
 
#include "PDF.h"
#include "FMUtils.h"
#include <cmath>
#include <iostream>
#include <fstream>
//#include <iostream>
#include <vector>
#include <Magick++.h>

using namespace std;
 

//***************************************************************************
//                        static functions
//***************************************************************************
string PDF::asText(IncMethod method)
{
  switch (method) {
    case imNone:   return "NONE"; 
    case imCube:   return "CUBE"; 
    case imSquare: return "SQUARE"; 
    case imCircle: return "CIRCLE"; 
    default: return "";
  }
}

PDF::IncMethod PDF::asIncMethod (string method)
{
  if (method.compare("NONE")==0) {
    return PDF::imNone;
  }
  if (method.compare("CUBE")==0) {
    return PDF::imCube;
  }
  if (method.compare("SQUARE")==0) {
    return PDF::imSquare;
  }
  if (method.compare("CIRCLE")==0) {
    return PDF::imCircle;
  }
  return PDF::imNone;
}

//***************************************************************************
//                         Constructors / Destructors
//***************************************************************************
PDF::PDF(unsigned int imgWidth, unsigned int imgHeight, unsigned int minScale, unsigned int maxScale, PdfParams pdfParams, std::string groundtruthfile)
{
#ifdef TESTHACK
  minScale=TESTHACK;
#endif
  //cout << "1" << flush;
  params = pdfParams;
  unsigned int objInitValue = params.initValue;
  if (groundtruthfile != "")
    objInitValue = 0;
  unsigned int expInitValue = params.maxValue;
  if (expInitValue == 0)
    expInitValue = 100;
  unsigned int samInitValue = (objInitValue*expInitValue)/100;
  
  resolution = 1;
  while (max(imgWidth,imgHeight) > (resolution*MAX_PDF_SIZE))
    resolution *= 2;
  //cout << "pdfres: " << resolution;
  xMax = imgWidth/resolution;    // for indexing the pdf use 0..xMax-(sMin+currScale)
  yMax = imgHeight/resolution;   // for indexing the pdf use 0..yMax-(sMin+currScale)
  sMax = min(min(imgWidth,imgHeight),maxScale)/resolution;
  sMin = min(minScale,maxScale)/resolution;
  sRange = sMax-sMin; // for indexing the pdf use 0..sRange
  //cout << endl << "Image: " << imgWidth << "x" << imgHeight << "  -->  " << xMax << "x" << yMax << flush;

  // ******** allocate Memory and initialize arrays ********
  //cout << "2" << flush;
  explorationDistribution = new unsigned int** [sRange+1]; //0..sRange
  objectDistribution = new unsigned int** [sRange+1]; //0..sRange
  samplingDistribution = new unsigned int** [sRange+1]; //0..sRange
  sumTable     = new unsigned int*  [sRange+1]; //0..sRange
  sumsumTable  = new unsigned int   [sRange+1]; //0..sRange
  sumsumsum = 0;
  for(unsigned int s=0; s<=sRange; s++) {
    sumsumTable[s] = 0;
    unsigned int xSize = xMax-(sMin+s)+1;
    sumTable[s]     = new unsigned int  [xSize]; //0..xMax-sMin-s
    explorationDistribution[s] = new unsigned int* [xSize]; //0..xMax-sMin-s
    objectDistribution[s] = new unsigned int* [xSize]; //0..xMax-sMin-s
    samplingDistribution[s] = new unsigned int* [xSize]; //0..xMax-sMin-s
    for(unsigned int x=0; x<xSize; x++) {
      unsigned int ySize = yMax-(sMin+s)+1;
      explorationDistribution[s][x] = new unsigned int [ySize]; //0..yMax-sMin-s
      objectDistribution[s][x] = new unsigned int [ySize]; //0..yMax-sMin-s
      samplingDistribution[s][x] = new unsigned int [ySize]; //0..yMax-sMin-s
      for(unsigned int y=0; y<ySize; y++) {
        explorationDistribution[s][x][y] = expInitValue; // uniform distribution
        objectDistribution[s][x][y] = objInitValue; // uniform distribution
        samplingDistribution[s][x][y] = samInitValue; // uniform distribution
      }
      unsigned int ySum = ySize * samInitValue;
      sumTable[s][x] = ySum;
      sumsumTable[s] += ySum;
      sumsumsum += ySum;
    }
  }
  // ******** use groundtruth file to set up pdf values ********
  //cout << "3" << flush;
  relObjSize = 0.0;
  unsigned int objCount = 0;
  if (groundtruthfile != "") {
    objInitValue = params.initValue;
    samInitValue = (objInitValue*expInitValue)/100;
    //cout << "setting up pdf with groundtruth file" << endl;
    ifstream inFile (groundtruthfile.c_str());
    if (!inFile) {
      cerr << "Unable to open groundtruthfile";
    } else {
      string s;
      while (!inFile.eof())
      {
        getline (inFile,s);
        if (s.length() > 7) {  //at least 4 numbers + 3 spaces inbetween
          vector<std::string> params;
          //cout << endl << "  processing " << s << flush;
          try {
            explode(s, " ", &params);
          } catch (int e) {
            continue;
          }
          int p0; int p1; int p2; int p3;
          from_string<int>(p0, params[0], std::dec);
          from_string<int>(p1, params[1], std::dec);
          from_string<int>(p2, params[2], std::dec);
          from_string<int>(p3, params[3], std::dec);
          //cout << endl << "    after split: " << p0 << " " << p1 << " " << p2 << " " << p3 << flush;
          p0 /= (int)resolution;
          p1 /= (int)resolution;
          p2 /= (int)resolution;
          p3 /= (int)resolution;
          //cout << endl << "    after resolution: " << p0 << " " << p1 << " " << p2 << " " << p3 << flush;
          p0 = min(max(0,p0),(int)xMax-1);
          p1 = min(max(0,p1),(int)yMax-1);
          p2 = max(min((int)xMax,p2),p0+(int)sMin);
          p3 = max(min((int)yMax,p3),p1+(int)sMin);
          //cout << endl << "    after minmax: " << p0 << " " << p1 << " " << p2 << " " << p3 << flush;
          if ((p2 > (int)xMax) || (p3 > (int)yMax))
            continue;
          // p2 = p2 + p0 + 1; // if coordinates are not given in absolute values (p1x, p1y, p2x, p2y)
          // p3 = p3 + p1 + 1; // then calculate absolute coordinates by adding width
          unsigned int w = p2 - p0;
          unsigned int h = p3 - p1;
          relObjSize += w*h;
          objCount++;
          unsigned int sLimit = min(sMax,min(w,h))-sMin;
          //cout << endl << "    sLimit: " << sLimit;
          for(unsigned int s=0; s<=sLimit; s++) {
            unsigned int xLimit = p2-(sMin+s)+1;
            unsigned int yLimit = p3-(sMin+s)+1;
            //cout << endl << "      xLimit: " << xLimit << "  yLimit: " << yLimit;
            for(unsigned int x=p0; x<xLimit; x++) {
              unsigned int ySum = 0;
              for(unsigned int y=p1; y<yLimit; y++) {
                if (objectDistribution[s][x][y] == 0) {
                  objectDistribution[s][x][y] = objInitValue; // uniform distribution
                  samplingDistribution[s][x][y] = samInitValue; // uniform distribution
                  ySum += samInitValue;
                }
              }
              sumTable[s][x] += ySum;
              sumsumTable[s] += ySum;
              sumsumsum += ySum;
            }
          } //-end for each scale s
        } //-end if line-length>7
      } //-end for each line
      if (objCount > 0) {
        relObjSize /= objCount; // result: in the average area of 
        relObjSize /= (xMax*yMax);
      }
    } //-end inFile opening sucessfull
  } //-end groundtruthfile != ""
  //if (sumsumsum < 1)
  //cout << "sum of PDF:" << sumsumsum << "...";
  // ******** if groundtruth information resultet in a zero pdf, initialize with 1 ********
  //cout << "4" << flush;
  if (sumsumsum == 0) {
    //cout << "5" << flush;
    objInitValue = params.initValue;
    samInitValue = (objInitValue*expInitValue)/100;
    for(unsigned int s=0; s<=sRange; s++) {
      unsigned int xSize = xMax-(sMin+s)+1;
      for(unsigned int x=0; x<xSize; x++) {
        unsigned int ySize = yMax-(sMin+s)+1;
        for(unsigned int y=0; y<ySize; y++) {
          objectDistribution[s][x][y] = objInitValue; // uniform distribution
          samplingDistribution[s][x][y] = samInitValue; // uniform distribution
        }
        sumTable[s][x] = ySize*samInitValue;
        sumsumTable[s] += sumTable[s][x];
        sumsumsum += sumsumTable[s];
      }
    }
  }
  // ******** initialise the random generator ********
  //randomInit();
}


PDF::~PDF()
{
  // ******** free Memory of arrays ********
  for(unsigned int s=0; s<=sRange; s++) {
    unsigned int xSize = xMax-(sMin+s)+1;
    for(unsigned int x=0; x<xSize; x++) {
      delete [] explorationDistribution[s][x];
      delete [] objectDistribution[s][x];
      delete [] samplingDistribution[s][x];
    }
    delete [] explorationDistribution[s];
    delete [] objectDistribution[s];
    delete [] samplingDistribution[s];
    delete [] sumTable[s];
  }
  delete [] explorationDistribution;
  delete [] objectDistribution;
  delete [] samplingDistribution;
  delete [] sumTable;
  delete [] sumsumTable;
}

  
//***************************************************************************
//                       public Methods
//***************************************************************************
/*!
  \brief method to draw a random point in scale space

  \param xPos pointer to the x-position (in range 0..width-1)
  \param yPos pointer to the z-position (in range 0..height-1)
  \param scale pointer to the scale (in range minScale..maxScale)
*/
void PDF::randomSample(unsigned int* xPos, unsigned int* yPos, unsigned int* scale)
{
  if (sumsumsum < 1)
    cout << "Sampling PDF is ZERO!!! I'm about to crash..." << flush;
  unsigned int randNumber = randomInRange(1, sumsumsum);
  
  // now sum up the numbers in distribution until the randNumber is reached
  // do this quickly using the sum-Tables
  unsigned int s=0;
  unsigned int x=0;
  unsigned int y=0;
  unsigned int currSum = 0;
  unsigned int currVal = 0;
  // fixing s
  while (currSum < randNumber) {  // &&(s<=sRange)
    currVal = sumsumTable[s];
    currSum += currVal;
    s++;
  }
  s--;
  //cout << endl << "select scale " << s << " (" << currSum << " , " << randNumber << ")";
  if (currSum == randNumber) {
    x = xMax-sMin-s;
    y = yMax-sMin-s;
    while (samplingDistribution[s][x][y] == 0) {
      if (y>0)
        y--;
      else {
        y = yMax-sMin-s;
        x--;
      }
    }
  }
  if (currSum > randNumber) {
    currSum -= currVal;
    // fixing x
    while (currSum < randNumber) {  // &&(x<xMax-sMin-s))
      currVal = sumTable[s][x];
      currSum += currVal;
      x++;
    }
    x--;
    //cout << endl << "select x " << x << " (" << currSum << " , " << randNumber << ")";
    if (currSum == randNumber) {
      y = yMax-sMin-s;
      while (samplingDistribution[s][x][y] == 0) {
        y--;
      }
    }
    if (currSum > randNumber) {
      currSum -= currVal;
      // fixing y
      while (currSum < randNumber) {  // &&(y<yMax-sMin-s))
        currVal = samplingDistribution[s][x][y];
        currSum += currVal;
        y++;
      }
      y--;
      //cout << endl << "select y " << y << " (" << currSum << " , " << randNumber << ")";
    }
  }
  *xPos = x * resolution;
  *yPos = y * resolution;
  *scale = (s+sMin) * resolution;
  //if ((x<114) || (y<22) ||(x+*scale>283) || (y+*scale>158)) 
  //  cout  << endl << " x,y / x,y: " << x << ","<< y << " / "<< x+*scale << ","<< y+*scale << "   randNumber:" << randNumber;
  increaseValue(s, x, y, -explorationDistribution[s][x][y], 0); // set value to 0, so that no window will be further sampled with exactely this parameters
}

void PDF::increaseValues(unsigned int scale, unsigned int xPos, unsigned int yPos, int explorationAmount, int objectAmount, double param1, double param2)
{
  scale /= resolution;
  scale -= sMin;
  xPos /= resolution;
  yPos /= resolution;
  //param1 /= resolution;
  //param2 /= resolution;
  switch (params.incMethod) {
    case imNone: ;
      break;
    case imCube: increaseCube(scale, xPos, yPos, explorationAmount, objectAmount, (unsigned int)(param1*min(xMax,yMax)/100)); 
      break;
    case imSquare: increaseSquare(scale, xPos, yPos, explorationAmount, objectAmount, (unsigned int)(param1*min(xMax,yMax)/100), (unsigned int)(param2*min(xMax,yMax)/100)); 
      break;
    case imCircle: increaseCircle(scale, xPos, yPos, explorationAmount, objectAmount, (unsigned int)(param1*min(xMax,yMax)/100));
      break;
  }
  //increaseValue(scale, xPos, yPos, -distribution[scale][xPos][yPos]+1); // set center to 1
}

void PDF::printIntoFile(string filename)
{
  string fn = filename;
  printIntoFile(fn.append("_obj.txt"), objectDistribution);
  fn = filename;
  printIntoFile(fn.append("_exp.txt"), explorationDistribution);
  fn = filename;
  printIntoFile(fn.append("_sam.txt"), samplingDistribution);
}

void PDF::printAsImage(std::string filename)
{
  string fn = filename;
  printAsImage(fn.append("_obj.jpg"), objectDistribution);
  fn = filename;
  printAsImage(fn.append("_exp.jpg"), explorationDistribution);
//   fn = filename;
//   printAsImage(fn.append("_sam.jpg"), samplingDistribution);
}

//***************************************************************************
//                        private Methods
//***************************************************************************
void PDF::increaseCube(unsigned int scale, unsigned int xPos, unsigned int yPos, int explorationAmount, int objectAmount, unsigned int radius)
{
  for (unsigned int s=max((unsigned int)0,scale-radius); (s<=scale+radius) && (s<=sRange); s++){
    //int pOffset = min(max(-s,(scale-s)/2),sRange-s);
    int pOffset = (scale-s)/2;
    for (unsigned int x=max((unsigned int)0,xPos-radius+pOffset); (x<=xPos+radius+pOffset) && (x<=xMax-sMin-s); x++){
      for (unsigned int y=max((unsigned int)0,yPos-radius+pOffset); (y<=yPos+radius+pOffset) && (y<=yMax-sMin-s); y++){
        increaseValue(s, x, y, explorationAmount, objectAmount);
      }
    }
  }
}

void PDF::increaseSquare(unsigned int scale, unsigned int xPos, unsigned int yPos, int explorationAmount, int objectAmount, unsigned int radius, unsigned int scaleradius)
{
  unsigned int leftBoundary = (unsigned int)max(0, (int)xPos-(int)radius);
  unsigned int rightBoundary = min(xMax, xPos+scale+sMin+radius);
  unsigned int topBoundary = (unsigned int)max(0, (int)yPos-(int)radius);
  unsigned int bottomBoundary = min(yMax, yPos+scale+sMin+radius);
  for (unsigned int s=max((unsigned int)0,scale-scaleradius); (s<=scale+scaleradius) && (s<=sRange); s++){
    for (unsigned int x=max((unsigned int)0,leftBoundary); (x+s+sMin<=rightBoundary); x++){
      for (unsigned int y=max((unsigned int)0,topBoundary); (y+s+sMin<=bottomBoundary); y++){
        increaseValue(s, x, y, explorationAmount, objectAmount);
      }
    }
  }
}

void PDF::increaseCircle(unsigned int scale, unsigned int xPos, unsigned int yPos, int explorationAmount, int objectAmount, unsigned int radius)
{
  for (unsigned int s=max(0,(int)scale-(int)radius); (s<=scale+radius) && (s<=sRange); s++){
    for (unsigned int x=max(0,(int)xPos-(int)radius); (x<=xPos+radius) && (x<=xMax-sMin-s); x++){
      for (unsigned int y=max(0,(int)yPos-(int)radius); (y<=yPos+radius) && (y<=yMax-sMin-s); y++){
        if (sqrt((s-scale)*(s-scale)+(x-xPos)*(x-xPos)+(y-yPos)*(y-yPos)) <= radius)
          increaseValue(s, x, y, explorationAmount, objectAmount);
      }
    }
  }
}


void PDF::increaseValue(unsigned int s, unsigned int x, unsigned int y, int explorationAmount, int objectAmount)
{
  if (params.maxValue > 0) {
    explorationAmount = min(explorationAmount, (int)params.maxValue-(int)explorationDistribution[s][x][y]);   //do not increase it to value > max
    objectAmount = min(objectAmount, (int)params.maxValue-(int)objectDistribution[s][x][y]);   //do not increase it to value > max
  }
  explorationAmount = max(explorationAmount, -((int)explorationDistribution[s][x][y])); //do not decrease it to value < 0
  objectAmount = max(objectAmount, -((int)objectDistribution[s][x][y])); //do not decrease it to value < 0
  explorationDistribution[s][x][y] += explorationAmount;
  objectDistribution[s][x][y] += objectAmount;
  int samplingAmount = max(1,(int)(explorationDistribution[s][x][y]*objectDistribution[s][x][y]/100)) - samplingDistribution[s][x][y]; //do not decrease it to value < 1
  samplingDistribution[s][x][y] += samplingAmount;
  sumTable[s][x] += samplingAmount;
  sumsumTable[s] += samplingAmount;
  sumsumsum += samplingAmount;
}

void PDF::printIntoFile(string filename, unsigned int ***distribution)
{
  ofstream outFile;
  outFile.open(filename.c_str(), ios::out);
  for(unsigned int s=0; s<=sRange; s++) {
    outFile << endl << "scale " << s+sMin << endl;
    unsigned int xSize = xMax-(sMin+s)+1;
    for(unsigned int x=0; x<xSize; x++) {
      unsigned int ySize = yMax-(sMin+s)+1;
      for(unsigned int y=0; y<ySize; y++)
        outFile << distribution[s][x][y] << "  ";
      outFile << endl;
    }
  }
  outFile.close();
}

void PDF::printAsImage(string filename, unsigned int ***distribution)
{
  // transform the distribution into 2D
  double dist2D[xMax][yMax];
  double count2D[xMax][yMax];
  for(unsigned int x=0; x<xMax; x++) {
    for(unsigned int y=0; y<yMax; y++) {
      dist2D[x][y] = 0;
      count2D[x][y] = 0;
		}
  }
  // slow method, but creates smoother images:
/*  for(unsigned int s=0; s<=sRange; s++) {
    unsigned int scale = sMin+s;
    unsigned int xSize = xMax-scale+1;
    for(unsigned int x=0; x<xSize; x++) {
      unsigned int ySize = yMax-scale+1;
      for(unsigned int y=0; y<ySize; y++) {
        // for each pixel in the current window increase the 2D distribution
			  for (unsigned int x2=x; x2<x+scale; x2++){
          for (unsigned int y2=y; y2<y+scale; y2++){
            dist2D[x2][y2] = dist2D[x2][y2] + (double)distribution[s][x][y];
            count2D[x2][y2]++;
          }
        }
      }
    }
  }*/
  // faster method, but creates fractional images:
  for (unsigned int x2=0; x2<xMax; x2++){
    for (unsigned int y2=0; y2<yMax; y2++){
      for(unsigned int s=0; s<=sRange; s++) {
        unsigned int dx = (s+sMin)/2;
        unsigned int x = x2-dx;
        unsigned int y = y2-dx;
        unsigned int xSize = xMax-(sMin+s)+1;
        unsigned int ySize = yMax-(sMin+s)+1;
        if ((x>=0) && (x<xSize) && (y>=0) && (y<ySize)) {
          dist2D[x2][y2] = dist2D[x2][y2] + (double)distribution[s][x][y];
          count2D[x2][y2]++;
        }
      }
    }
  }
  double cmax = 0;
  for(unsigned int x=0; x<xMax; x++) {
    for(unsigned int y=0; y<yMax; y++) {
      dist2D[x][y] /= count2D[x][y];
      if (dist2D[x][y] > cmax)
        cmax = dist2D[x][y];
		}
  }
  // now create the output image
  //image(xMax+1,yMax+1)
  Magick::Image outimage( Magick::Geometry(xMax, yMax), "white" );
  for(unsigned int x=0; x<xMax; x++) {
    for(unsigned int y=0; y<yMax; y++) {
      outimage.pixelColor(x,y,Magick::ColorGray(dist2D[x][y]/cmax));
		}
  }
  //outimage.draw(Magick::DrawablePointSize(18)); // set the point size
  //outimage.draw(Magick::DrawableStrokeColor("black"));
  //outimage.draw(Magick::DrawableFillColor("red"));
  //outimage.draw(Magick::DrawableText(10, 10, to_String(cmax)));
  outimage.write(filename.c_str());
}

// --- none ---
