/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    \file SvmClassifier.cpp

    $Author: Frank Moosmann $ ( <moosmann@mrt.uka.de>)
    $Date: 2006/01/14 $
*///////////////////////////////////////////////////////////////////////////
 

#include "SvmClassifier.h"
#include "ClassLabel.h"

#include <iostream>
#include <istream>
#include <fstream>
#include <cmath>
#include <list>

using namespace std;
 
/***************************************************************************
                         Constructors / Destructors
 ***************************************************************************/
/*!
  \brief Method for creating a SVM from the examples

  This method creates a SVM model based on the given feature list
  \param featVecList the list with FeatureVectors and ClassLabels
  \param params structure containing the parameters
*/
SvmClassifier::SvmClassifier(list<FeatureVector*> *featVecList, svm_parameter *params)
{
  ClassLabelHistogram totalHistogram = ClassLabelHistogram(featVecList);
  totalHistogram.getListOfClassLabels(&classLabelList);                 // set up classLabelList
  convertFeatureVectors(featVecList);
  svm = svm_train(&prob, params);
/*	cout << endl << "class labels:";
	int i = 0;
  for(list<ClassLabel>::iterator p=classLabelList.begin(); p!=classLabelList.end(); ++p) {
		cout << endl << *p << "\t" << svm->label[i];
		i++;
	}
	cout << endl;*/
}

/*!
  \brief Method for creating a SVM from the given SVM-Problem

  This method creates a SVM model based on the given problem data-structure
  \param problem the given SVM-problem
  \param classHistogram a histogram of the classLabels
  \param params structure containing the parameters
*/
SvmClassifier::SvmClassifier(struct svm_problem *problem, ClassLabelHistogram *classHistogram, svm_parameter *params)
{
  prob.l = 0;
  prob.y = 0;
  prob.x = 0;
  classHistogram->getListOfClassLabels(&classLabelList);                 // set up classLabelList
  svm = svm_train(problem, params);
}

/*!
  \brief Method to read SvmClassifier from a file

  \param filename file to create classifier from
*/
SvmClassifier::SvmClassifier(string filename)
{
  prob.l = 0;
  prob.y = 0;
  prob.x = 0;
  svm = 0;
  ifstream inFile (filename.c_str());
  // read treecount
  if (! inFile.eof() )
  {
    string s;
    getline(inFile,s);
    //cout << s << endl;
    char classlabels [255];
    if (sscanf(s.c_str(), "SVMClassifier%s", classlabels) == 1) {
      string remainingLabels = classlabels;
      while (remainingLabels.length() > 0) {
        int label;
        int count = sscanf(remainingLabels.c_str(), ";%d%s", &label, classlabels);
        if (count >= 1)
          classLabelList.push_back(label);
        if (count == 2)
          remainingLabels = classlabels;
        else
          remainingLabels = "";
      }
    } // scanf first line
  } // !eof
  if (! inFile.eof() ) // read svm model
  {
    string s;
    getline(inFile,s);
    char fn [256];
    if (sscanf(s.c_str(), "svmmodel=%s", fn) == 1) {
      svm = svm_load_model(fn);
    }
  }
  inFile.close();
}


SvmClassifier::~SvmClassifier ( )
{
  if (svm != 0) {
    svm_destroy_model(svm);
  }
  if (prob.x != 0) {
    for (int i=0; i<prob.l; i++) {
      delete [] prob.x[i];
    }
    delete [] prob.x;
  }
  if (prob.y != 0) {
    delete [] prob.y;
  }
}



/***************************************************************************
                         public Methods
 ***************************************************************************/

void SvmClassifier::writeToFile(std::string filename)
{
  ofstream outfile;
  outfile.open(filename.c_str(), ios::out);
  outfile << "SVMClassifier";
  for(list<ClassLabel>::iterator p=classLabelList.begin(); p!=classLabelList.end(); ++p)
    outfile << ";" << *p;
  outfile << endl;
  if (svm != 0) {
    filename.append(".svmmodel");
    svm_save_model(filename.c_str(), svm);
    outfile << "svmmodel=" << filename << endl;
  }
  outfile.close();
}

/*!
  \brief Method for classifying a new example

  This method classifies the new example based on the SVM model
  created with the method learnFromExamples().
  \param featureVec the list with FeatureVectors of the example
  \param totalVotes pointer to the histrogram where the votes for all windows will be stored (optional)
  \param windowVotes pointer to the histrogram where the votes for the class will be stored (optional).
         In this case the result will be equivalent to totalVotes
  \return the estimated class
*/
ClassLabel SvmClassifier::classify (list<FeatureVector*> *featureVec, ClassLabelHistogram *totalVotes, ClassLabelHistogram *windowVotes)
{
  //debug_out(1, "starting to classify...");
  unsigned int classCount = classLabelList.size();
  ClassLabel negativeClass = *(classLabelList.begin());
  confidenceValue = 0;
  ClassLabelHistogram* featureVotes = new ClassLabelHistogram();
  for(list<FeatureVector*>::const_iterator featI=featureVec->begin(); (featI!=featureVec->end()); ++featI)
  {
    FeatureVector *fVec = *featI;
    struct svm_node *fVecSvmNode;
    fVecSvmNode = new svm_node[fVec->getSize()+1]; // +1 because sparse list ends with additional entry "-1"
    convertFeatureVector(fVec, fVecSvmNode);
    ClassLabel featureClassLabel = (ClassLabel)svm_predict(svm, fVecSvmNode);
    //debug_out(6, "voting for class %i", classLabel);

    fVec->classLabel = featureClassLabel;
    featureVotes->increaseCountOfLabel(featureClassLabel);
    if (totalVotes != 0)
      totalVotes->increaseCountOfLabel(featureClassLabel);
    if (windowVotes != 0)
      windowVotes->increaseCountOfLabel(fVec->classLabel);
    if (classCount == 2) { // in two-class-case create statistic about confidence of decision
      double a[1];
      svm_predict_values(svm, fVecSvmNode, a);  //gives nr_class*(nr_class-1)/2 decision values
      if (featureClassLabel==negativeClass)
        a[0] = -abs(a[0]);
      else
        a[0] = +abs(a[0]);
      confidenceValue += a[0];
    }
    delete fVecSvmNode;
  }
  ClassLabel classifiedLabel = featureVotes->getRealLabelWithMaxOccurance();
  if (classCount == 2) { // in two-class-case create statistic about confidence of decision
    confidenceValue /= featureVec->size();
  }
  delete featureVotes;
  return classifiedLabel;
}


/*!
  \brief Method for getting the weights of each dimension calculated by the SVM

  \param w pointer to already allocated memory to store the results
  \param dim number of dimensions
*/
void SvmClassifier::getWeights(double** w, int dim)
{
  svm_get_w(svm, w, dim);
}



/***************************************************************************
                         private Methods
 ***************************************************************************/

/*!
  \brief Method for converting FeatureVectors into svm_problem

  ATTENTION: this method allocated memory for parameters "prob" and "x_space".
             memory must be deallocated later by using free(prob.x), free(prob.y) and free(x_space)

  \param featureVec the list with FeatureVectors
*/
void SvmClassifier::convertFeatureVectors (std::list<FeatureVector*>* featureVec) {
  /* struct svm_problem
    int l;
    double *y;
    struct svm_node **x;
  */
  int fCount = featureVec->size();
  int fullSize = (*(featureVec->begin()))->getSize();
  prob.l = fCount;
  prob.y = new double [fCount];
  prob.x = new svm_node* [fCount];
  int i = 0;
  for(list<FeatureVector*>::const_iterator featI=featureVec->begin(); featI!=featureVec->end(); ++featI) {
    FeatureVector* fVec = *featI;
    int compactSize = 0;
    for (int j=0; j<fullSize; j++) {
      if (fVec->getElement(j) != 0) {
        compactSize++;
      }
    }
    prob.y[i] = fVec->classLabel;
    prob.x[i] = new svm_node[compactSize+1];
    convertFeatureVector(fVec, prob.x[i]);
    i++;
  }
/*
  cout << endl << "start training with " << prob.l << " features";
  cout << endl << "class of feature 4: " << prob.y[4];
  cout << endl << "data  of feature 4: " << prob.x[4]->value;
  struct svm_node *x = prob.x[4];
  while (x->index != -1) {
    cout << endl << "    (" << x->index << "|" << x->value << ")";
    x++;
  }
*/
}

/*!
  \brief Method for converting one FeatureVector into a svm_node
  
  ASSUMPTION: enough memory for svm_nodes is already allocated (FEATURE_NUMBER+1)

  \param fVec pointer to the FeatureVector
  \param x pointer to the memory where the data will be stored (must already be allocated!!!)
 */
void SvmClassifier::convertFeatureVector (FeatureVector* fVec, struct svm_node *x) {
  /* struct svm_node
  int index;
  double value;
  */
  //cout << endl << "processing feature vector of size" << fVec->getSize() << flush;
  int fullSize = fVec->getSize();
  for (int i=0; i<fullSize; i++) {
    if (fVec->getElement(i) != 0) {
      x->index = i;
      x->value = fVec->getElement(i);
      x++;
    } else {
      //cout << " " << i << "X";
    }
  }
  x->index = -1;
}

